/* -*-C-*- ttanh.c */

#include "elefunt.h"

/*
#     program to test tanh
#
#     data required
#
#        none
#
#     subprograms required from this package
#
#        machar - an environmental inquiry program providing
#                 information on the floating-point arithmetic
#                 system.  note that the call to machar can
#                 be deleted provided the following five
#                 parameters are assigned the values indicated
#
#                 ibeta  - the radix of the floating-point system
#                 it     - the number of base-ibeta digits in the
#                          significand of a floating-point number
#                 minexp - the largest in magnitude negative
#                          integer such that float(ibeta)**minexp
#                          is a positive floating-point number
#                 xmin   - the smallest non-vanishing floating-point
#                          power of the radix
#                 xmax   - the largest finite floating-point no.
#
#        ran(k) - a function subprogram returning random real
#                 numbers uniformly distributed over (0,1)
#
#
#     standard fortran subprograms required
#
#         abs, alog, amax1, float, sqrt, tanh
#
#
#     latest revision - december 6, 1979
#
#     author - w. j. cody
#              argonne national laboratory
#
#*/

void
ttanh(VOID_ARG)
{
    int i,
        ibeta,
        iexp,
        irnd,
        it,
        j,
        k1,
        k2,
        k3,
        machep,
        maxexp,
        minexp,
        n,
        negep,
        ngrd;

    dp_t
        eps,
        epsneg,
        xmax,
        xmin;

    volatile dp_t 
	a,
        ait,
        albeta,
        b,
        beta,
        betap,
        c,
        d,
        del,
        expon,
        half,
        r6,
        r7,
        w,
        x,
        xl,
        xn,
        x1,
        y,
        z,
        zz;

    /*******************************************************************/

    (void)ranset(initseed());
    machar(&ibeta, &it, &irnd, &ngrd, &machep, &negep, &iexp, &minexp,
        &maxexp, &eps, &epsneg, &xmin, &xmax);

    beta = TO_FP_T(ibeta);
    albeta = ALOG(beta);
    ait = TO_FP_T(it);
    half = 0.5e0;
    a = 0.125e0;
    b = ALOG(3.0e0) * half;
    c = 1.2435300177159620805e-1;
    d = ALOG(2.0e0) + (ait + ONE) * ALOG(beta) * half;
    n = maxtest();
    xn = TO_FP_T(n);
    z = ZERO;

    /* random argument accuracy tests */

    for (j = 1; j <= 2; ++j)
    {
	k1 = 0;
	k3 = 0;
	x1 = ZERO;
	r6 = ZERO;
	r7 = ZERO;
	del = (b - a) / xn;
	xl = a;

	for (i = 1; i <= n; ++i)
	{
	    x = del * RAN() + xl;
	    z = TANH(x);
	    y = x - 0.125e0;
	    zz = TANH(y);
	    zz = (zz + c) / (ONE + c * zz);
	    w = ONE;
	    if (z != ZERO)
		w = (z - zz) / z;
	    if (w > ZERO)
		k1 = k1 + 1;
	    if (w < ZERO)
		k3 = k3 + 1;
	    w = ABS(w);
	    if (w > r6)
	    {
		r6 = w;
		x1 = x;
	    }
	    r7 = r7 + w * w;
	    xl = a + TO_FP_T(i) * del;	/* OLD: xl = xl + del (bad for large n) */
	}

	k2 = n - k3 - k1;
	r7 = SQRT(r7 / xn);
	(void)printf("1TEST OF TANH(X) VS (TANH(X-1/8)+TANH(1/8))/(1+TANH(X-1/8)TANH(1/8))\n\n");
	(void)printf("%7d RANDOM ARGUMENTS WERE TESTED FROM THE INTERVAL\n", n);
	(void)printf("      (%15.4e,%15.4e)\n\n\n", a, b);
	(void)printf(" TANH(X) WAS LARGER%6d TIMES,\n", k1);
	(void)printf("             AGREED%6d TIMES, AND\n", k2);
	(void)printf("        WAS SMALLER%6d TIMES.\n\n", k3);
	(void)printf(" THERE ARE%4d BASE%4d SIGNIFICANT DIGITS IN A FLOATING-POINT NUMBER\n\n",
	    it, ibeta);
	w = -999.0e0;
	if (r6 != ZERO)
	    w = ALOG(ABS(r6)) / albeta;
	(void)printf(" THE MAXIMUM RELATIVE ERROR OF%15.4e = %4d **%7.2f\n",
	    r6, ibeta, w);
	(void)printf("    OCCURRED FOR X =%17.6e\n", x1);
	w = AMAX1(ait + w, ZERO);
	(void)printf(
	    " THE ESTIMATED LOSS OF BASE%4d SIGNIFICANT DIGITS IS%7.2f\n\n\n",
	    ibeta, w);
	w = -999.0e0;
	if (r7 != ZERO)
	    w = ALOG(ABS(r7)) / albeta;
	(void)printf(" THE ROOT MEAN SQUARE RELATIVE ERROR WAS%15.4e = %4d **%7.2f\n",
	    r7, ibeta, w);
	w = AMAX1(ait + w, ZERO);
	(void)printf(
	    " THE ESTIMATED LOSS OF BASE%4d SIGNIFICANT DIGITS IS%7.2f\n\n\n",
	    ibeta, w);
	a = b + a;
	b = d;
    }

    /* special tests */

    (void)printf("1SPECIAL TESTS\n\n");
    (void)printf(" THE IDENTITY   TANH(-X) = -TANH(X)   WILL BE TESTED.\n\n\n");
    (void)printf("        X         F(X) + F(-X)\n\n");

    for (i = 1; i <= 5; ++i)
    {
	x = RAN();
	z = TANH(x) + TANH(-x);
	(void)printf("%15.7e%15.7e\n\n", x, z);
    }

    (void)printf(" THE IDENTITY TANH(X) = X , X SMALL, WILL BE TESTED.\n\n\n");
    (void)printf("        X         X - F(X)\n\n");
    betap = IPOW(beta, it);
    x = RAN() / betap;

    for (i = 1; i <= 5; ++i)
    {
	z = x - TANH(x);
	(void)printf("%15.7e%15.7e\n\n", x, z);
	x = x / beta;
    }

    (void)printf(" THE IDENTITY TANH(X) = 1 , X LARGE, WILL BE TESTED.\n\n\n");
    (void)printf("        X         1 - F(X)\n\n");
    x = d;
    b = 4.0e0;

    for (i = 1; i <= 5; ++i)
    {
	z = (TANH(x) - half) - half;
	(void)printf("%15.7e%15.7e\n\n", x, z);
	x = x + b;
    }

    (void)printf(" TEST OF UNDERFLOW FOR VERY SMALL ARGUMENT.\n\n");
    expon = TO_FP_T(minexp) *0.75e0;
    x = POW(beta, expon);
    z = TANH(x);
    (void)printf("       TANH(%13.6e) =%13.6e\n", x, z);
    (void)printf(" THE FUNCTION TANH WILL BE CALLED WITH THE ARGUMENT%15.7e\n", xmax);
    z = TANH(xmax);
    (void)printf("       TANH(%13.6e) =%13.6e\n", xmax, z);
    (void)printf(" THE FUNCTION TANH WILL BE CALLED WITH THE ARGUMENT%15.7e\n", xmin);
    z = TANH(xmin);
    (void)printf("       TANH(%13.6e) =%13.6e\n", xmin, z);
    x = ZERO;
    (void)printf(" THE FUNCTION TANH WILL BE CALLED WITH THE ARGUMENT%15.7e\n", x);
    z = TANH(x);
    (void)printf("       TANH(%13.6e) =%13.6e\n", x, z);
    (void)printf(" THIS CONCLUDES THE TESTS\n");
}
